package net.projectmonkey.object.mapper.construction.type;

import net.projectmonkey.object.mapper.analysis.result.PropertyPath;
import net.projectmonkey.object.mapper.analysis.result.PropertyPathElement;
import net.projectmonkey.object.mapper.construction.PopulationContext;
import net.projectmonkey.object.mapper.context.ExecutionContext;
import net.projectmonkey.object.mapper.util.CollectionUtil;
import net.projectmonkey.object.mapper.util.MapUtil;
import net.projectmonkey.object.mapper.util.TypeResolver;
import org.apache.commons.lang3.ArrayUtils;

import java.lang.reflect.GenericArrayType;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.lang.reflect.TypeVariable;
import java.util.*;

import static net.projectmonkey.object.mapper.util.TypeResolver.Unknown;

/**
 * @author Andy Moody
 */
public class GenericTypeUtils
{

	public static GenericTypeUtils INSTANCE = new GenericTypeUtils();

	private GenericTypeUtils(){}

	public Class<?> resolveActualDestinationType(final PopulationContext<?, ?> context, final PropertyPath sourcePath, final PropertyPath destPath)
	{
		Class<?> destinationType = destPath.getProperty().getType();
		Type genericType = destPath.getProperty().getGenericType();
		if(genericType instanceof TypeVariable)
		{
			Class<?> toReturn = destinationType;
			Map<TypeVariable<?>, Class<?>> destinationTypeArguments = context.getDestinationTypeArguments();
			if(MapUtil.hasElements(destinationTypeArguments) && destinationTypeArguments.containsKey(genericType))
			{
				toReturn = destinationTypeArguments.get((TypeVariable) genericType);
			}
			destinationType = toReturn;
		}
		return destinationType;
	}

	public TypeAndClass[] getArgumentTypes(final PopulationContext<?, ?> context)
	{
		Type genericType = context.getDestinationGenericType();
		return getArgumentTypes(context, genericType);
	}

	/**
	 * Resolves a single argument types for a sub property. E.G. given an object Structure like:
	 * class Entity<K, V> {
	 *    List<K> list;
	 * }
	 *
	 * @param destinationType
	 * @param context
	 * @return
	 */
	public TypeAndClass getArgumentType(final Class<?> destinationType, final PopulationContext<?, ?> context)
	{
		TypeAndClass[] types = getArgumentTypes(context);
		TypeAndClass argumentType = new TypeAndClass(Unknown.class);
		if (types != null)
		{
			if (types.length != 1)
			{
				throw new IllegalArgumentException("Expected 1 type argument on generic type " + destinationType.getName() + " but found " + types.length);
			}
			argumentType = types[0];
		}
		return argumentType;
	}

	public static void addDestinationTypeArguments(final PopulationContext<?, ?> context)
	{
		Map<Class<?>, GenericTypeResolver> genericTypeResolvers = ExecutionContext.getConfiguration().getGenericTypeResolvers();
		Object source = context.getSource();
		if(source != null)
		{
			Class<?> sourceType = source.getClass();
			Class<?> destinationType = context.getDestinationType();
			TypeVariable<? extends Class<?>>[] destinationTypeParameters = destinationType.getTypeParameters();
			Map<TypeVariable<?>, Set<TypeVariable<?>>> topLevelTypeVars = TypeResolver.getTopLevelTypeVars(destinationType);
			if(genericTypeResolvers.containsKey(sourceType) && destinationTypeParameters.length > 0)
			{
				List<Class<?>> destinationTypeArguments = genericTypeResolvers.get(sourceType).getDestinationTypeArguments(source);
				Map<TypeVariable<?>, Class<?>> destinationTypesByVariable = new LinkedHashMap<TypeVariable<?>, Class<?>>();
				for (int i = 0; i < destinationTypeParameters.length; i++)
				{
					if(destinationTypeArguments.size() > i)
					{
						Set<TypeVariable<?>> typeVariables = topLevelTypeVars.get(destinationTypeParameters[i]);
						for (TypeVariable<?> typeVariable : typeVariables)
						{
							destinationTypesByVariable.put(typeVariable, destinationTypeArguments.get(i));
						}
					}
				}
				context.addDestinationTypeArguments(destinationTypesByVariable);
			}
		}
	}

	public static Map<TypeVariable<?>, Class<?>> migrateDestinationTypeArguments(final PropertyPathElement destPath, final PopulationContext parentContext)
	{
		Map<TypeVariable<?>, Class<?>> toReturn = new HashMap<TypeVariable<?>, Class<?>>();

		Map<TypeVariable<?>, Class> parentDestinationTypeArguments = parentContext.getDestinationTypeArguments();
		Type genericType = destPath.getGenericType();
		Class<?> destinationType = destPath.getType();

		addMappedTypeArguments(toReturn, parentDestinationTypeArguments, genericType, destinationType);
		return toReturn;
	}

	private static void addMappedTypeArguments(final Map<TypeVariable<?>, Class<?>> toReturn, final Map<TypeVariable<?>, Class> parentDestinationTypeArguments, final Type genericType,
											   final Class<?> destinationType)
	{
		if(genericType instanceof ParameterizedType)
		{
			ParameterizedType parameterizedType = (ParameterizedType) genericType;
			TypeVariable<? extends Class<?>>[] typeParameters = destinationType.getTypeParameters();
			if(parameterizedType.getRawType() instanceof Class)
			{
				typeParameters = ((Class) parameterizedType.getRawType()).getTypeParameters();
			}
			addMappedTypeArguments(toReturn, parentDestinationTypeArguments, parameterizedType, destinationType, typeParameters);
		}
		else if(genericType instanceof GenericArrayType)
		{
			GenericArrayType genericArrayType = (GenericArrayType) genericType;
			addMappedTypeArguments(toReturn, parentDestinationTypeArguments, genericArrayType.getGenericComponentType(), destinationType.getComponentType());
		}
		else if(genericType instanceof TypeVariable)
		{
			addMappedType(toReturn, parentDestinationTypeArguments, (TypeVariable<?>) genericType, genericType);
		}
	}

	private static void addMappedType(final Map<TypeVariable<?>, Class<?>> toReturn, final Map<TypeVariable<?>, Class> parentDestinationTypeArguments,
									  final TypeVariable<?> typeParameter, final Type actualTypeArgument)
	{
		if(parentDestinationTypeArguments.containsKey(actualTypeArgument))
		{
			toReturn.put(typeParameter, parentDestinationTypeArguments.get(actualTypeArgument));
		}
		else if (actualTypeArgument instanceof ParameterizedType)
		{
			ParameterizedType parameterizedType = (ParameterizedType) actualTypeArgument;
			if(parameterizedType.getRawType() instanceof Class)
			{
				Class<?> destinationType = (Class<?>) parameterizedType.getRawType();
				TypeVariable<?>[] typeParameters = ((Class) parameterizedType.getRawType()).getTypeParameters();
				addMappedTypeArguments(toReturn, parentDestinationTypeArguments, parameterizedType, destinationType, typeParameters);
			}
		}
	}

	private static void addMappedTypeArguments(final Map<TypeVariable<?>, Class<?>> toReturn, final Map<TypeVariable<?>, Class> parentDestinationTypeArguments,
											   final ParameterizedType parameterizedType, final Class<?> destinationType, final TypeVariable<?>[] typeParameters)
	{
		Type[] actualTypeArguments = parameterizedType.getActualTypeArguments();
		Map<TypeVariable<?>, Set<TypeVariable<?>>> topLevelTypeVars = TypeResolver.getTopLevelTypeVars(destinationType);
		for(int i = 0; i < actualTypeArguments.length; i++)
		{
			Type subTypeArgument = actualTypeArguments[i];
			for (TypeVariable<?> subTypeParameter : topLevelTypeVars.get(typeParameters[i]))
			{
				addMappedType(toReturn, parentDestinationTypeArguments, subTypeParameter, subTypeArgument);
			}
		}
	}

	private TypeAndClass[] getArgumentTypes(final PopulationContext<?, ?> context, final Type genericType)
	{
		TypeVariable[] typeParameters = new TypeVariable[0];
		Type[] actualTypeArguments = new Type[0];
		TypeAndClass[] toReturn = new TypeAndClass[0];

		if(genericType instanceof Class)
		{
			typeParameters = ((Class) genericType).getTypeParameters();
			actualTypeArguments = ((Class) genericType).getTypeParameters();
		}
		else if(genericType instanceof TypeVariable)
		{
			typeParameters = new TypeVariable[]{(TypeVariable) genericType};
			actualTypeArguments = typeParameters;
		}
		else if(genericType instanceof GenericArrayType)
		{
			GenericArrayType genericArrayType = (GenericArrayType) genericType;
			Type genericComponentType = genericArrayType.getGenericComponentType();
			if(genericComponentType instanceof TypeVariable)
			{
				toReturn = getArgumentTypes(context, genericComponentType);
			}
			else if (genericComponentType instanceof ParameterizedType)
			{
				Type rawType = ((ParameterizedType) genericComponentType).getRawType();
				if(rawType instanceof Class)
				{
					toReturn = new TypeAndClass[]{new TypeAndClass(genericComponentType, (Class<?>) rawType)};
				}
			}
		}
		else if(genericType instanceof ParameterizedType)
		{
			ParameterizedType parameterizedType = (ParameterizedType) genericType;
			Type rawType = parameterizedType.getRawType();
			actualTypeArguments = parameterizedType.getActualTypeArguments();
			if(rawType instanceof Class)
			{
				typeParameters = ((Class) rawType).getTypeParameters();
			}
			else
			{
				typeParameters = context.getDestinationType().getTypeParameters();
			}
		}
		if(ArrayUtils.isEmpty(toReturn))
		{
			toReturn = new TypeAndClass[typeParameters.length];
			Map<TypeVariable<?>, Set<TypeVariable<?>>> topLevelTypeVars = TypeResolver.getTopLevelTypeVars(context.getDestinationType());
			for (int i = 0; i < typeParameters.length; i++)
			{
				TypeVariable typeParameter = typeParameters[i];
				Map<TypeVariable<?>, Class<?>> destinationTypeArguments = context.getDestinationTypeArguments();
				if(topLevelTypeVars.containsKey(typeParameter) )
				{
					for (TypeVariable<?> topLevelTypeParameter : topLevelTypeVars.get(typeParameter))
					{
						if(destinationTypeArguments.containsKey(topLevelTypeParameter))
						{
							toReturn[i] = createTypeAndClass(actualTypeArguments, topLevelTypeParameter, destinationTypeArguments, i);
							break;
						}
					}
				}
				else if(destinationTypeArguments.containsKey(typeParameter))
				{
					TypeAndClass typeAndClass = createTypeAndClass(actualTypeArguments, typeParameter, destinationTypeArguments, i);
					toReturn[i] = typeAndClass;
				}
			}
			populateUnmappedDestinationTypes(context, toReturn, genericType, actualTypeArguments);
		}

		return toReturn;
	}

	private TypeAndClass createTypeAndClass(final Type[] actualTypeArguments, final TypeVariable typeParameter, final Map<TypeVariable<?>, Class<?>> destinationTypeArguments, final int index)
	{
		Class<?> clazz = destinationTypeArguments.get(typeParameter);
		Type type = actualTypeArguments.length > index ? actualTypeArguments[index] : clazz;
		return new TypeAndClass(type, clazz);
	}

	private void populateUnmappedDestinationTypes(final PopulationContext<?, ?> context, final TypeAndClass[] toReturn,
												  final Type genericType, Type[] actualTypeArguments)
	{
		List<Integer> unresolvedPositions = new ArrayList<Integer>();
		for (int i = 0; i < toReturn.length; i++)
		{
			if(toReturn[i] == null)
			{
				toReturn[i] = new TypeAndClass(Unknown.class);
				unresolvedPositions.add(i);
			}
		}
		if(CollectionUtil.hasElements(unresolvedPositions))
		{
			Class<?> parentDestinationType = context.getParentDestinationType();
			if(parentDestinationType != null)
			{
				Class<?>[] arguments = TypeResolver.resolveArguments(genericType, parentDestinationType);
				for (Integer unresolvedPosition : unresolvedPositions)
				{
					if(arguments != null && arguments.length > unresolvedPosition)
					{
						Class<?> argument = arguments[unresolvedPosition];
						Type type = actualTypeArguments.length > unresolvedPosition ? actualTypeArguments[unresolvedPosition] : argument;
						toReturn[unresolvedPosition] = new TypeAndClass(type, argument);
					}
				}
			}
		}
	}

	public static class TypeAndClass
	{
		private Type type;
		private Class<?> clazz;

		public TypeAndClass(final Type type, final Class<?> clazz)
		{
			this.type = type;
			this.clazz = clazz;
		}

		public TypeAndClass(final Class<?> clazz)
		{
			this(clazz, clazz);
		}

		public Type getType()
		{
			return type;
		}

		public Class<?> getClazz()
		{
			return clazz;
		}
	}

}
