package net.projectmonkey.object.mapper.util;

import java.lang.ref.Reference;
import java.lang.ref.WeakReference;
import java.lang.reflect.*;
import java.util.*;

/*
 *
 *  * Copyright 2012 the original author or authors.
 *  *
 *  * Licensed under the Apache License, Version 2.0 (the "License");
 *  * you may not use this file except in compliance with the License.
 *  * You may obtain a copy of the License at
 *  *
 *  *      http://www.apache.org/licenses/LICENSE-2.0
 *  *
 *  * Unless required by applicable law or agreed to in writing, software
 *  * distributed under the License is distributed on an "AS IS" BASIS,
 *  * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  * See the License for the specific language governing permissions and
 *  * limitations under the License.
 *
 */

/**
 * Enhanced type resolution utilities. Based on org.springframework.core.GenericTypeResolver.
 * <p/>
 * Adapted from ModelMapper(modelmapper.org)
 *
 * @author Jonathan Halterman
 * @author Andy Moody
 */
public final class TypeResolver
{

	/**
	 * An unknown type.
	 */
	public static final class Unknown
	{
		private Unknown()
		{
		}
	}

	private TypeResolver()
	{
	}

	/**
	 * Cache of type variable/argument pairs
	 */
	private static final Map<Class<?>, Reference<Map<TypeVariable<?>, Type>>> typeVariableCache = Collections.synchronizedMap(new WeakHashMap<Class<?>, Reference<Map<TypeVariable<?>, Type>>>());
	private static final Map<Class<?>, Reference<Map<TypeVariable<?>, Set<TypeVariable<?>>>>> typLevelTypeVarsCache =
			Collections.synchronizedMap(new WeakHashMap<Class<?>, Reference<Map<TypeVariable<?>, Set<TypeVariable<?>>>>>());


	/**
	 * Resolves the type argument for the {@code genericType} using type variable information from the
	 * {@code sourceType}. If {@code genericType} is an instance of class, then {@code genericType} is
	 * returned. If no arguments can be resolved then {@code Unknown.class} is returned.
	 *
	 * @param genericType to resolve upwards from
	 * @param targetType  to resolve arguments for
	 * @return type argument for {@code initialType} else {@code null} if no type arguments are
	 *         declared
	 * @throws IllegalArgumentException if more or less than one type argument is resolved for the
	 *                                  give types
	 */
	public static Class<?> resolveArgument(Type genericType, Class<?> targetType)
	{
		Class<?>[] arguments = resolveArguments(genericType, targetType);
		if (arguments == null)
		{
			return Unknown.class;
		}

		if (arguments.length != 1)
		{
			throw new IllegalArgumentException("Expected 1 type argument on generic type "
													   + targetType.getName() + " but found " + arguments.length);
		}

		return arguments[0];
	}

	/**
	 * Returns an array of raw classes representing type arguments for the {@code targetType} resolved
	 * upwards from the {@code initialType}. Arguments for {@code targetType} that cannot be resolved
	 * to a Class are returned as {@code Unknown.class}. If no arguments can be resolved then
	 * {@code null} is returned.
	 *
	 * @param initialType to resolve upwards from
	 * @param targetType  to resolve arguments for
	 * @return array of raw classes representing type arguments for {@code initialType} else
	 *         {@code null} if no type arguments are declared
	 */
	public static <T, I extends T> Class<?>[] resolveArguments(Class<I> initialType, Class<T> targetType)
	{
		return resolveArguments(resolveGenericType(initialType, targetType), initialType);
	}

	/**
	 * Resolves the arguments for the {@code genericType} using the type variable information for the
	 * {@code targetType}. Returns {@code null} if {@code genericType} is not parameterized or if
	 * arguments cannot be resolved.
	 */
	public static Class<?>[] resolveArguments(Type genericType, Class<?> targetType)
	{
		Class<?>[] result = null;

		if (genericType instanceof ParameterizedType)
		{
			ParameterizedType paramType = (ParameterizedType) genericType;
			Type[] arguments = paramType.getActualTypeArguments();
			result = new Class[arguments.length];
			for (int i = 0; i < arguments.length; i++)
			{
				result[i] = resolveClass(arguments[i], targetType);
			}
		}
		else if (genericType instanceof TypeVariable)
		{
			result = new Class[1];
			result[0] = resolveClass(genericType, targetType);
		}

		return result;
	}

	/**
	 * Resolves the generic Type for the {@code targetType} by walking the type hierarchy upwards from
	 * the {@code initialType}.
	 */
	public static Type resolveGenericType(Type initialType, Class<?> targetType)
	{
		Class<?> rawType;
		if (initialType instanceof ParameterizedType)
		{
			rawType = (Class<?>) ((ParameterizedType) initialType).getRawType();
		}
		else
		{
			rawType = (Class<?>) initialType;
		}

		if (targetType.equals(rawType))
		{
			return initialType;
		}

		Type result;
		if (targetType.isInterface())
		{
			for (Type superInterface : rawType.getGenericInterfaces())
			{
				if (superInterface != null && !superInterface.equals(Object.class))
				{
					if ((result = resolveGenericType(superInterface, targetType)) != null)
					{
						return result;
					}
				}
			}
		}

		Type superType = rawType.getGenericSuperclass();
		if (superType != null && !superType.equals(Object.class))
		{
			if ((result = resolveGenericType(superType, targetType)) != null)
			{
				return result;
			}
		}

		return null;
	}

	/**
	 * Resolves the raw class for the given {@code genericType}, using the type variable information
	 * from the {@code targetType}.
	 */
	public static Class<?> resolveClass(Type genericType, Class<?> targetType)
	{
		if (genericType instanceof Class)
		{
			return (Class<?>) genericType;
		}
		else if (genericType instanceof ParameterizedType)
		{
			return resolveClass(((ParameterizedType) genericType).getRawType(), targetType);
		}
		else if (genericType instanceof GenericArrayType)
		{
			GenericArrayType arrayType = (GenericArrayType) genericType;
			Class<?> compoment = resolveClass(arrayType.getGenericComponentType(), targetType);
			return Array.newInstance(compoment, 0).getClass();
		}
		else if (genericType instanceof TypeVariable)
		{
			TypeVariable<?> variable = (TypeVariable<?>) genericType;
			genericType = getTypeVariableMap(targetType).get(variable);
			genericType = genericType == null ? resolveBound(variable) : resolveClass(genericType,
																					  targetType);
		}

		return genericType instanceof Class ? (Class<?>) genericType : Unknown.class;
	}

	public static Map<TypeVariable<?>, Type> getTypeVariableMap(final Class<?> targetType)
	{
		Reference<Map<TypeVariable<?>, Type>> ref = typeVariableCache.get(targetType);
		Map<TypeVariable<?>, Type> map = ref != null ? ref.get() : null;

		if (map == null)
		{
			map = new HashMap<TypeVariable<?>, Type>();

			// Populate interfaces
			buildTypeVariableMap(targetType.getGenericInterfaces(), map);

			// Populate super classes and interfaces
			Type genericType = targetType.getGenericSuperclass();
			Class<?> type = targetType.getSuperclass();
			while (type != null && !Object.class.equals(type))
			{
				if (genericType instanceof ParameterizedType)
				{
					buildTypeVariableMap((ParameterizedType) genericType, map);
				}
				buildTypeVariableMap(type.getGenericInterfaces(), map);

				genericType = type.getGenericSuperclass();
				type = type.getSuperclass();
			}

			// Populate enclosing classes
			type = targetType;
			while (type.isMemberClass())
			{
				genericType = type.getGenericSuperclass();
				if (genericType instanceof ParameterizedType)
				{
					buildTypeVariableMap((ParameterizedType) genericType, map);
				}

				type = type.getEnclosingClass();
			}
			typeVariableCache.put(targetType, new WeakReference<Map<TypeVariable<?>, Type>>(map));
		}

		return map;
	}

	/**
	 * @param initialType
	 * @return the top level type variables for each of the type variables in the initialType
	 */
	public static Map<TypeVariable<?>, Set<TypeVariable<?>>> getTopLevelTypeVars(final Class<?> initialType)
	{
		Reference<Map<TypeVariable<?>, Set<TypeVariable<?>>>> ref = typLevelTypeVarsCache.get(initialType);
		Map<TypeVariable<?>, Set<TypeVariable<?>>> map = ref != null ? ref.get() : null;

		if(map == null)
		{
			map = new HashMap<TypeVariable<?>, Set<TypeVariable<?>>>();
			Map<TypeVariable<?>, TypeVariable<?>> currentVarToInitialVarMap = new HashMap<TypeVariable<?>, TypeVariable<?>>();
			for (TypeVariable<?> typeVar : initialType.getTypeParameters())
			{
				map.put(typeVar, new HashSet<TypeVariable<?>>());
				currentVarToInitialVarMap.put(typeVar, typeVar);
			}

			Type genericSuperclass = initialType.getGenericSuperclass();
			Class<?> clazz = initialType.getSuperclass();

			while (clazz != null && !Object.class.equals(clazz))
			{
				buildTypeHierarchy(currentVarToInitialVarMap, new Type[]{genericSuperclass});

				genericSuperclass = clazz.getGenericSuperclass();
				clazz = clazz.getSuperclass();
			}

			for (Map.Entry<TypeVariable<?>, TypeVariable<?>> entry : currentVarToInitialVarMap.entrySet())
			{
				TypeVariable topLevelVar = entry.getKey();
				TypeVariable initialVar = entry.getValue();
				if(map.containsKey(initialVar))
				{
					map.get(initialVar).add(topLevelVar);
				}
			}
			typLevelTypeVarsCache.put(initialType, new WeakReference<Map<TypeVariable<?>, Set<TypeVariable<?>>>>(map));
		}

		return map;
	}

	private static void buildTypeHierarchy(final Map<TypeVariable<?>, TypeVariable<?>> currentVarToInitialVarMap, final Type[] types)
	{
		for (Type type : types)
		{
			if(type instanceof ParameterizedType)
			{
				Type[] actualTypeArguments = ((ParameterizedType) type).getActualTypeArguments();
				Type rawType = ((ParameterizedType) type).getRawType();
				if(rawType instanceof Class)
				{
					TypeVariable[] typeVariables = ((Class) rawType).getTypeParameters();
					Set<TypeVariable<?>> toRemove = new HashSet<TypeVariable<?>>();
					for (int i = 0; i < actualTypeArguments.length; i++)
					{
						Type actualTypeArgument = actualTypeArguments[i];
						if(currentVarToInitialVarMap.containsKey(actualTypeArgument))
						{
							currentVarToInitialVarMap.put(typeVariables[i], currentVarToInitialVarMap.get(actualTypeArgument));
							toRemove.add((TypeVariable<?>) actualTypeArgument);
						}
					}

					for (TypeVariable<?> typeVariable : toRemove)
					{
						currentVarToInitialVarMap.remove(typeVariable);
					}
				}
			}
		}
	}

	/**
	 * Populates the {@code map} with with variable/argument pairs for the given {@code types}.
	 */
	static void buildTypeVariableMap(final Type[] types, final Map<TypeVariable<?>, Type> map)
	{
		for (Type type : types)
		{
			if (type instanceof ParameterizedType)
			{
				ParameterizedType parameterizedType = (ParameterizedType) type;
				buildTypeVariableMap(parameterizedType, map);
				Type rawType = parameterizedType.getRawType();
				if (rawType instanceof Class)
				{
					buildTypeVariableMap(((Class<?>) rawType).getGenericInterfaces(), map);
				}
			}
			else if (type instanceof Class)
			{
				buildTypeVariableMap(((Class<?>) type).getGenericInterfaces(), map);
			}
		}
	}

	/**
	 * Populates the {@code typeVariableMap} with type arguments and parameters for the given
	 * {@code type}.
	 */
	private static void buildTypeVariableMap(ParameterizedType type,
											 Map<TypeVariable<?>, Type> typeVariableMap)
	{
		if (type.getRawType() instanceof Class)
		{
			TypeVariable<?>[] typeVariables = ((Class<?>) type.getRawType()).getTypeParameters();
			Type[] typeArguments = type.getActualTypeArguments();

			for (int i = 0; i < typeArguments.length; i++)
			{
				TypeVariable<?> variable = typeVariables[i];
				Type typeArgument = typeArguments[i];

				if (typeArgument instanceof Class)
				{
					typeVariableMap.put(variable, typeArgument);
				}
				else if (typeArgument instanceof GenericArrayType)
				{
					typeVariableMap.put(variable, typeArgument);
				}
				else if (typeArgument instanceof ParameterizedType)
				{
					typeVariableMap.put(variable, typeArgument);
				}
				else if (typeArgument instanceof TypeVariable)
				{
					TypeVariable<?> typeVariableArgument = (TypeVariable<?>) typeArgument;
					Type resolvedType = typeVariableMap.get(typeVariableArgument);
					if (resolvedType == null)
					{
						resolvedType = resolveBound(typeVariableArgument);
					}
					typeVariableMap.put(variable, resolvedType);
				}
			}
		}
	}

	/**
	 * Resolves the first bound for the {@code typeVariable}, returning {@code Unknown.class} if none
	 * can be resolved.
	 */
	public static Type resolveBound(TypeVariable<?> typeVariable)
	{
		Type[] bounds = typeVariable.getBounds();
		if (bounds.length == 0)
		{
			return Unknown.class;
		}

		Type bound = bounds[0];
		if (bound instanceof TypeVariable)
		{
			bound = resolveBound((TypeVariable<?>) bound);
		}

		return bound == Object.class ? Unknown.class : bound;
	}

}
