package net.projectmonkey.object.mapper.util;

import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Member;
import java.lang.reflect.Method;
import java.util.*;

/*
 *
 *  * Copyright 2012 the original author or authors.
 *  *
 *  * Licensed under the Apache License, Version 2.0 (the "License");
 *  * you may not use this file except in compliance with the License.
 *  * You may obtain a copy of the License at
 *  *
 *  *      http://www.apache.org/licenses/LICENSE-2.0
 *  *
 *  * Unless required by applicable law or agreed to in writing, software
 *  * distributed under the License is distributed on an "AS IS" BASIS,
 *  * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  * See the License for the specific language governing permissions and
 *  * limitations under the License.
 *
 */

/**
 * <p/>
 * Utilities for working with types.
 *
 * Adapted from ModelMapper(modelmapper.org)
 *
 * @author Jonathan Halterman
 * @author Andy moody
 */
public final class Types
{
	private static Class<?> JAVASSIST_PROXY_FACTORY_CLASS;
	private static Method JAVASSIST_IS_PROXY_CLASS_METHOD;
	private static Class[] KNOWN_TYPES;
	private static Map<Class<?>, Boolean> knownTypesCache;

	static
	{
		try
		{
			KNOWN_TYPES = new Class[]{String.class, Map.class, Enum.class, Date.class, Calendar.class, java.sql.Date.class, Number.class};
			knownTypesCache = new HashMap<Class<?>, Boolean>();
			JAVASSIST_PROXY_FACTORY_CLASS = Types.class.getClassLoader().loadClass(
					"javassist.util.proxy.ProxyFactory");
			JAVASSIST_IS_PROXY_CLASS_METHOD = JAVASSIST_PROXY_FACTORY_CLASS.getMethod("isProxyClass",
																					  new Class<?>[]{Class.class});
		}
		catch(Exception ignore)
		{
		}
	}

	private Types()
	{
	}

	/**
	 * Returns the proxied type, if any, else returns the given {@code type}.
	 */
	@SuppressWarnings("unchecked")
	public static <T> Class<T> deProxy(Class<?> type)
	{
		// Ignore JDK proxies
		if(type.isInterface())
		{
			return (Class<T>) type;
		}

		// CGLib
		if(type.getName().contains("$$EnhancerByCGLIB$$") || type.getName().contains("$$EnhancerByMockitoWithCGLIB$$"))
		{
			return (Class<T>) type.getSuperclass();
		}

		// Javassist
		try
		{
			if(JAVASSIST_IS_PROXY_CLASS_METHOD != null && JAVASSIST_IS_PROXY_CLASS_METHOD != null
					&& (Boolean) JAVASSIST_IS_PROXY_CLASS_METHOD.invoke(null, type))
			{
				return (Class<T>) type.getSuperclass();
			}
		}
		catch(Exception ignore)
		{
		}

		return (Class<T>) type;
	}

	public static boolean isKnownType(Class<?> clazz)
	{
		boolean knownType;
		if(knownTypesCache.containsKey(clazz))
		{
			knownType = knownTypesCache.get(clazz);
		}
		else
		{
			knownType = Primitives.isPrimitiveOrWrapper(clazz) || Iterables.isIterable(clazz);

			if(!knownType)
			{
				for (int i = 0; i < KNOWN_TYPES.length && !knownType; i++)
				{
					if(KNOWN_TYPES[i].isAssignableFrom(clazz))
					{
						knownType = true;
					}
				}
			}
			knownTypesCache.put(clazz, knownType);
		}
		return knownType;
	}

	public static boolean isArrayOfType(Class<?> clazz, Class<?>...types)
	{
		boolean toReturn = false;
		List<Class<?>> typeList = Arrays.asList(types);
		if(clazz.isArray())
		{
			toReturn = typeList.contains(clazz.getComponentType());
		}
		return toReturn;
	}

	/**
	 * Returns a simplified String representation of the {@code type}.
	 */
	public static String toString(Class<?> type)
	{
		return type.getName();
	}

	/**
	 * Returns a simplified String representation of the {@code member}.
	 */
	public static String toString(Member member)
	{
		if(member instanceof Method)
		{
			return member.getDeclaringClass().getName() + "." + member.getName() + "()";
		}
		else if(member instanceof Field)
		{
			return member.getDeclaringClass().getName() + "." + member.getName();
		}
		else if(member instanceof Constructor)
		{
			return member.getDeclaringClass().getName() + ".<init>()";
		}
		return null;
	}

}